/**
  *  \file arch/win32-bcc55/platform.cc
  */

#include <windows.h>
#include "arch/win32-bcc55/platform.h"
#include "io/dirs.h"
#include "cpluslib/utf8.h"
#include "arch/win32/win32.h"

const char Platform::LINE_SEPARATOR[] = "\r\n";
const char Platform::PATH_SEPARATOR;

/** Get name of program directory. */
static string_t
getProgDirName()
{
    enum { N = 1024 };
    if (win32::hasUnicodeSupport()) {
        wchar_t buffer[N];
        int len = GetModuleFileNameW(NULL, buffer, N);
        return Directory::getDirName(convertWcharToUtf8(buffer, len));
    } else {
        char buffer[N];
        int len = GetModuleFileNameA(NULL, buffer, N);
        buffer[len] = 0;
        return Directory::getDirName(win32::convertFromANSI(buffer));
    }
}

void
Platform::initPlatform()
{
    /* Prevent Windows from displaying a dialog box when we access a
       disk drive which is not ready. */
    SetErrorMode(SEM_FAILCRITICALERRORS);
}

string_t
Platform::getDataDirectory()
{
    return Directory::makeFileName(getProgDirName(), "specs");
}

string_t
Platform::getResourceDirectory()
{
    return Directory::makeFileName(getProgDirName(), "resource");
}

string_t
Platform::getLanguageCode()
{
    /* According to Win32.hlp:
       # LOCALE_SABBREVLANGNAME:
       #    Abbreviated name of the language, created by taking the
       #    2-letter language abbreviation from the ISO Standard 639 and
       #    adding a third letter, as appropriate, to indicate the
       #    sublanguage.
       So, when we get `enu', we make `en_u' from it. Under Unix,
       the relevant code would be `en_US', but we only use the part
       before the underscore anyway. */

    /* "Language codes and scripts":
      Language/Locale           Code    Script        Unix equivalent
      Greek                     ELL     Other         el_GR.iso8859-7
      Russian                   RUS     Cyrillic      ru_RU
      Turkish                   TRK     Latin 2       tr_TR
      Polish                    PLK     Latin 2       pl_PL
      Czech                     CSY     Latin 2       cs_CZ
      Slovak                    SKY     Latin 2       sk_SK
      Hungarian                 HUN     Latin 2       hu_HU
      Danish                    DAN     Latin 1       da_DK
      Dutch (Standard)          NLD     Latin 1       nl (??)
      Belgian (Flemish)         NLB     Latin 1       be (??)
      English (American)        ENU     Latin 1       en_US
      English (British)         ENG     Latin 1       en_GB
      English (Australian)      ENA     Latin 1
      English (Canadian)        ENC     Latin 1
      English (New Zealand)     ENZ     Latin 1
      English (Ireland)         ENI     Latin 1
      Finnish                   FIN     Latin 1       fi_FI
      French (Standard)         FRA     Latin 1       fr_FR
      French (Belgian)          FRB     Latin 1
      French (Canadian)         FRC     Latin 1
      French (Swiss)            FRS     Latin 1
      German (Standard)         DEU     Latin 1       de_DE
      German (Swiss)            DES     Latin 1       de_CH
      German (Austrian)         DEA     Latin 1       de_AT
      Icelandic                 ISL     Latin 1       is_IS
      Italian (Standard)        ITA     Latin 1       it_IT
      Italian (Swiss)           ITS     Latin 1
      Norwegian (Bokmal)        NOR     Latin 1       no_NO
      Norwegian (Nynorsk)       NON     Latin 1       nn_NO
      Portuguese (Brazilian)    PTB     Latin 1       pt_BR
      Portuguese (Standard)     PTG     Latin 1       pt_PT
      Swedish                   SVE     Latin 1       sv_SE
      Spanish (Std/Trad)        ESP     Latin 1       es_ES
      Spanish (Mexican)         ESM     Latin 1
      Spanish (Modern)          ESN     Latin 1 */
    
    char buffer[10];
    int n = GetLocaleInfo(LOCALE_USER_DEFAULT, LOCALE_SABBREVLANGNAME,
                          buffer, sizeof buffer);
    if (n <= 1) {
        return string_t();
    } else {
        --n;
        if (n <= 2)
            return string_t(buffer, n);
        else
            return string_t(buffer, 2) + "_" + string_t(buffer+2, n-2);
    }
}

static void
appendCString(std::vector<char*>& out, const string_t& accum)
{
    char* str = new char[accum.size()+1];
    accum.copy(str, accum.size(), 0);
    str[accum.size()] = 0;
    out.push_back(str);
}

static bool
tryExpandWildcard(std::vector<char*>& out, const string_t& pattern)
{
    bool did = false;
    if (win32::hasUnicodeSupport()) {
        win32::WStr uniPattern;
        win32::convertToUnicode(uniPattern, pattern, true);

        size_t uniPos = uniPattern.size();
        while (uniPos > 0 && (uniPattern[uniPos-1] != L'/' && uniPattern[uniPos-1] != L'\\' && uniPattern[uniPos-1] != L':')) {
            --uniPos;
        }

        WIN32_FIND_DATAW data;
        HANDLE h = FindFirstFileW(&uniPattern[0], &data);
        if (h != INVALID_HANDLE_VALUE) {
            do {
                if (wcscmp(data.cFileName, L".") != 0 && wcscmp(data.cFileName, L"..") != 0) {
                    win32::WStr uniCombined(&uniPattern[0], &uniPattern[uniPos]);
                    for (size_t i = 0; data.cFileName[i] != 0; ++i) {
                        uniCombined.push_back(data.cFileName[i]);
                    }
                    uniCombined.push_back(0);
                    appendCString(out, win32::convertFromUnicode(&uniCombined[0]));
                    did = true;
                }
            } while (FindNextFileW(h, &data));
            FindClose(h);
        }
    } else {
        string_t ansiPattern = win32::convertToANSI(pattern);
        string_t::size_type ansiPos = ansiPattern.find_last_of("/\\:");
        if (ansiPos == string_t::npos) {
            ansiPos = 0;
        } else {
            ++ansiPos;
        }
        WIN32_FIND_DATAA data;
        HANDLE h = FindFirstFileA(ansiPattern.c_str(), &data);
        if (h != INVALID_HANDLE_VALUE) {
            do {
                if (strcmp(data.cFileName, ".") != 0 && strcmp(data.cFileName, "..") != 0) {
                    string_t ansiCombined(ansiPattern, 0, ansiPos);
                    ansiCombined += data.cFileName;
                    appendCString(out, win32::convertFromANSI(ansiCombined.c_str()));
                    did = true;
                }
            } while (FindNextFileA(h, &data));
            FindClose(h);
        }
    }
    return did;
}                              

/** Build argument array. This builds an UTF-8-encoded argv array from a single
    Unicode command line string. The out parameter can be 0 to just count the
    strings instead of generating them; use this to allocate the output buffer.
    \param out [out] argv strings go here.
    \param in  [in] Unicode command line
    \return Number of argv strings produced */
static void
buildArgs(std::vector<char*>& out, const wchar_t* in)
{
    while (1) {
        // Skip whitespace
        while (*in && (*in == L'\t' || *in == L' '))
            ++in;
        if (*in == 0)
            break;

        // Parse argument
        string_t accum;
#if 0
        if (*in == L'"') {
            // quoted string
            ++in;
            while (1) {
                wchar_t ch = *in++;
                if (ch == L'"') {
                    if (*in == L'"') {
                        // quote
                        ++in;
                    } else {
                        // end
                        break;
                    }
                }
                appendUtf8(accum, ch);
            }
        } else {
            // single word
            while (*in && *in != L'\t' && *in != L' ')
                appendUtf8(accum, *in++);
        }
#else
        bool hadWild = false;
        bool isQuoted = false;
        while (wchar_t ch = *in) {
            if (ch == L'"') {
                // Quote
                if (isQuoted) {
                    // Within a quote. Double-quote produces single quote output.
                    // Otherwise, just end the quote.
                    if (in[1] == L'"') {
                        ++in;
                        appendUtf8(accum, ch);
                    } else {
                        isQuoted = false;
                    }
                } else {
                    // Starting a quote
                    isQuoted = true;
                }
            } else if (ch == L'\\' && isQuoted && in[1] == L'\"') {
                // Cygwin quote
                appendUtf8(accum, in[1]);
                ++in;
            } else if (ch == L':' || ch == L'\\' || ch == L'/') {
                // Path delimiter. Forget that we had a wildcard.
                hadWild = false;
                appendUtf8(accum, ch);
            } else if (ch == L'*' || ch == L'?') {
                // Wildcard.
                if (!isQuoted) {
                    hadWild = true;
                }
                appendUtf8(accum, ch);
            } else if (ch == L' ' || ch == L'\t') {
                if (!isQuoted) {
                    break;
                } else {
                    appendUtf8(accum, ch);
                }
            } else {
                appendUtf8(accum, ch);
            }
            ++in;
        }
#endif

        // Remember it
        if (hadWild && tryExpandWildcard(out, accum)) {
            // ok
        } else {
            appendCString(out, accum);
        }
    }
}

/** Convert argument strings. Since the argv strings received by
    main() do not contain characters outside the system codepage, this
    uses the Unicode command line provided by the API instead of the
    passed-in argv. */
char**
Platform::convertArgumentsToUtf8(char** argv)
{
    try {
        /* Fetch command line. This Unicode function is supported on Win95. */
        LPWSTR cl = GetCommandLineW();

        /* Build argv strings. CommandLineToArgvW is not supported on Win95,
           and there is no CommandLineToArgvA, so we use our own routine. */
        std::vector<char*> args;
        buildArgs(args, cl);
        size_t nargs = args.size();
        char** results(new char*[nargs+1]);
        std::copy(args.begin(), args.end(), results);
        results[nargs] = 0;

        return results;
    }
    catch (...) {
        MessageBox(0, "Unable to initialize application.", argv[0], MB_ICONSTOP | MB_OK);
        std::exit(1);
    }
}

/** Free argument strings. This frees the result produced by
    Platform::convertArgumentsToUtf8. */
void
Platform::freeArguments(char** argv)
{
    for (int i = 0; argv[i] != 0; ++i)
        delete[] argv[i];
    delete[] argv;
}
