/**
  *  \file client/tiles/planetgrowth.cc
  *  \brief Planet Growth Tile
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/tiles/planetgrowth.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/planetpred.h"
#include "util/unicodechars.h"
#include "util/misc.h"
#include "util/strtab.h"
#include "ui/util.h"
#include "game/pref.h"
#include "game/config.h"

namespace {
    bool hasExperience(const GPlanet& planet, const GGameConfig& config)
    {
        return config.NumExperienceLevels() > 0
            && planet.unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpPoints)) >= 0;
    }
    
    int getNumLines(const GPlanet& planet, const GGameConfig& config)
    {
        int result = 1;
        if (planet.getNativeRace() > 0) {
            ++result;
        }
        if (hasExperience(planet, config)) {
            ++result;
        }
        return result;
    }
}

static ConfigIntOption cfg_showDiffs(getUserPreferences(), "Tax.PredictRelative", ValueBoolParser::instance);

/** Constructor. */
WPlanetGrowthTile::WPlanetGrowthTile(const GPlanet& planet)
    : WTile(font_sizes[FONT_NORMAL].scaledBy(30, getNumLines(planet, config))),
      planet(planet),
      eff()
{ }

/** Set planetary structures for prediction.
    \param str structure type
    \param n amount */
void
WPlanetGrowthTile::setStructures(TPlanetaryStructure str, int n)
{
    planet.setStructures(str, mp16_t(n));
    drawWidget();
}

/** Set native tax for prediction.
    \param n tax rate */
void
WPlanetGrowthTile::setNativeTax(int n)
{
    planet.setNativeTax(mp16_t(n));
    drawWidget();
}

/** Set colonist tax for prediction.
    \param n tax rate */
void
WPlanetGrowthTile::setColonistTax(int n)
{
    planet.setColonistTax(mp16_t(n));
    drawWidget();
}

/** Set planet effectors. Those affect prediction.
    \param eff effectors */
void
WPlanetGrowthTile::setEffectors(GPlanetEffectors eff)
{
    this->eff = eff;
    drawWidget();
}

void
WPlanetGrowthTile::drawData(GfxCanvas& can)
{
    /* Initial status */
    GPlanet work(planet);
    int oldColTax = work.getColonistTax();
    int oldNatTax = work.getNativeTax();
    bool natives = work.getNatives() > 0;
    bool exp = hasExperience(work, config);
    int expPoints = work.unit_scores.getScore(planet_score_definitions.lookupScore(ScoreId_ExpPoints));

    /* Graphics status */
    GfxRect size = getInnerSize(0, 0);
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setColor(UIColor::tc_Static);

    int x  = size.x;
    int y  = size.y;
    int y2 = y + font_heights[FONT_NORMAL];
    int y3 = natives ? y2 + font_heights[FONT_NORMAL] : y2;

    /* Labels */
    const char*const tx[] = { N_("Colonists:"), N_("Natives:"), N_("Experience:") };
    outText(ctx, x, y, _(tx[0]));
    if (natives)
        outText(ctx, x, y2, _(tx[1]));
    if (exp)
        outText(ctx, x, y3, _(tx[2]));

    /* Values */
    int colWidth = 6*font_ems[FONT_NORMAL];
    int dx = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(tx, 3)) + colWidth;
    ctx.setTextAlign(2, 0);
    ctx.setSolidBackground();
    while (dx < size.w) {
        /* Compute next turn */
        int32_t oldCol = work.getColonists();
        int32_t oldNat = work.getNatives();
        int32_t oldExp = expPoints;
        computeTurn(work, eff);
        if (exp) {
            expPoints += config.EPPlanetAging();

            int happy = work.getColonistHappiness();
            if (work.getNatives() > 0) {
                happy = std::min(happy, int(work.getNativeHappiness()));
            }
            if (happy > 0) {
                expPoints += config.EPPlanetGovernment() * happy / 100;
            }
        }
        int32_t newCol = work.getColonists();
        int32_t newNat = work.getNatives();
        int32_t newExp = expPoints;

        /* Show it */
        ctx.setColor(work.getColonistTax() == oldColTax ? UIColor::tc_Static : UIColor::tc_Faded);
        showItem(ctx, x+dx, y, oldCol, newCol);
        if (natives) {
            ctx.setColor(work.getNativeTax() == oldNatTax ? UIColor::tc_Static : UIColor::tc_Faded);
            showItem(ctx, x+dx, y2, oldNat, newNat);
        }
        if (exp) {
            ctx.setColor(UIColor::tc_Static);
            showExp(ctx, x+dx, y3, oldExp, newExp);
        }
        dx += colWidth;
    }
}

bool
WPlanetGrowthTile::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key && event.key.code == 'r') {
        setSelectedWidget();
        cfg_showDiffs.set(!cfg_showDiffs());
        drawWidget();
        return true;
    } else {
        return WTile::handleEvent(event, second_pass);
    }
}

void
WPlanetGrowthTile::showItem(GfxContext& ctx, int x, int y, int32_t oldPop, int32_t newPop)
{
    int colWidth = 6*font_ems[FONT_NORMAL];
    if (newPop == 0 && (oldPop == 0 || !cfg_showDiffs())) {
        outTextF(ctx, x, y, colWidth, UTF_MIDDLE_DOT);
    } else if (cfg_showDiffs()) {
        string_t diffs = clansToString(std::abs(newPop - oldPop));
        if (newPop < oldPop)
            diffs.insert(0, "-");
        if (newPop > oldPop)
            diffs.insert(0, "+");
        outTextF(ctx, x, y, colWidth, diffs);
    } else {
        outTextF(ctx, x, y, colWidth, clansToString(newPop));
    }
}

void
WPlanetGrowthTile::showExp(GfxContext& ctx, int x, int y, int32_t oldExp, int32_t newExp)
{
    int colWidth = 6*font_ems[FONT_NORMAL];
    string_t text;
    if (newExp == 0) {
        text = UTF_MIDDLE_DOT;
    } else if (cfg_showDiffs()) {
        text = numToString(std::abs(newExp - oldExp));
        if (newExp > oldExp)
            text.insert(0, "+");
    } else {
        text = numToString(newExp);
    }

    int level = 0;
    while (level < config.NumExperienceLevels() && oldExp >= config.ExperienceLevels(level+1)) {
        ++level;
    }
    if (level < config.NumExperienceLevels() && newExp >= config.ExperienceLevels(level+1)) {
        text.insert(0, UTF_UP_ARROW " ");
    }

    outTextF(ctx, x, y, colWidth, text);
}
