/**
  *  \file game/global.cc
  *  \brief Global Turn Data
  *
  *  This contains most of the GGameTurn class.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/global.h"
#include "game/ship.h"
#include "util/nls.h"
#include "util/console.h"
#include "util/except.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "game/host.h"
#include "cpluslib/misc.h"
#include "game/racename.h"
#include "game/gamedirs.h"
#include "game/team.h"
#include "game/msgglobal.h"
#include "game/fleet.h"
#include "game/classicvcr.h"
#include "game/palliance.h"
#include "game/talliance.h"

/** Global turn instance. */
Ptr<GGameTurn> current_turn_internal;
Ptr<GGameTurn> display_turn_internal;

GProxyObjectType current_ship_type;             // temporary!
GProxyObjectType current_planet_type;           // temporary!
GProxyObjectType current_base_type;             // temporary!
GProxyObjectType current_ion_type;              // temporary!
GProxyObjectType current_minefield_type;        // temporary!
GProxyObjectType current_anyplanet_type;        // temporary!
GProxyObjectType current_anyship_type;          // temporary!
GProxyObjectType current_histship_type;         // temporary!
GProxyObjectType current_ufo_type;              // temporary;
GProxyObjectType current_fleet_type;            // temporary!

GSimpleObjectSelection current_ship_selection;             // temporary!
GSimpleObjectSelection current_planet_selection;           // temporary!
GSimpleObjectSelection current_base_selection;             // temporary!
GSimpleObjectSelection current_ion_selection;              // temporary!
GSimpleObjectSelection current_minefield_selection;        // temporary!
GSimpleObjectSelection current_histship_selection;         // temporary!
GSimpleObjectSelection current_ufo_selection;              // temporary;
GSimpleObjectSelection current_fleet_selection;            // temporary!

/******************************* GGameTurn *******************************/

/*! \class GGameTurn
    This class contains everything that makes up a game turn: things
    that change each turn. There's one global instance of this class
    which is used by most code; some special-purpose code will work
    on different instances.

    This contains many functions that otherwise would be global,
    hence the huuuuuuuge interface.

    During a run, we expect all planet/starbase objects to be present,
    but possibly empty. To not have to do this during static
    initialisation (which isn't possible with Ptr<> anyway), one
    must call init() before loading the turn. Basically, do
    \code
        try {
            GGameTurn& t = getCurrentTurn();
            t.init();
            // load game...
            t.postprocess(...);
        }
        catch(std::exception& e) {
            // ...
        }
    \endcode
*/

/** Constructor. Creates a blank GGameTurn. You must call init()
    before you can use it. */
GGameTurn::GGameTurn()
    : playing_set(), data_set(),
      database_turn(0)
{
    init();
}

/** Destructor. Frees all memory. */
GGameTurn::~GGameTurn()
{
    done();
}

/** Create all data. */
void
GGameTurn::init()
{
    std::fill_n(player_data, countof(player_data), (Player*) 0);
    vcrs = new GClassicVcrDatabase();
}

/** Clear all data. */
void
GGameTurn::done()
{
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        delete player_data[i-1];
        player_data[i-1] = 0;
    }
    vcrs = 0;
}


/** Add database turn. This remembers the newest database turn seen. */
void
GGameTurn::addDatabaseTurn(int n)
{
    if (n > database_turn)
        database_turn = n;
}

/** Get database turn. */
int
GGameTurn::getDatabaseTurn() const
{
    return database_turn;
}

GGameTurn::Player&
GGameTurn::createPlayerData(int player)
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    if (!player_data[player-1])
        player_data[player-1] = new Player(player);
    return *player_data[player-1];
}

const GGameTurn::Player&
GGameTurn::getPlayerDataNC(int player) const
{
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    return *player_data[player-1];
}

GGameTurn::Player&
GGameTurn::getPlayerData(int player)
{
    // FIXME: not much difference to createPlayerData
    ASSERT(player > 0 && player <= NUM_PLAYERS);
    return player_data[player-1] ? *player_data[player-1] : createPlayerData(player);
}


/** Set GEN data for player.
    \param player player
    \param gen parsed GENx.DAT for that player. */
void
GGameTurn::setGen(int player, const GGen& gen)
{
    getPlayerData(player).gen = gen;
    if (!dat.getTurnNumber()) {
        dat.setTurnNumber(gen.getTurnNumber());
        dat.setTimestamp(gen.getTimestamp());
        dis.setTurnNumber(gen.getTurnNumber());
        dis.setTimestamp(gen.getTimestamp());
    }
}

/** Set that we have data for the specified player. */
void
GGameTurn::setHaveData(int player)
{
    createPlayerData(player);
    data_set |= player;
}

/** Postprocess after loading. */
void
GGameTurn::postprocess(GObject::Playability playability)
{
    playing_set = data_set;

    host.finish();

    // At this time, we have loaded util.dat, and know the host version.
    // Create the alliance thing for all players
    // (if util.dat reading already created one, this is a no-op).
    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (data_set.contains(i)) {
            if (host.isPHost()) {
                GPHostAllianceHandler::processVersion(getCommands(i), getAlliances(i), host.host_version);
            } else {
                GTHostAllianceHandler::processVersion(getCommands(i), getAlliances(i), host.host_version);
            }
        }
    }

    console.write(LOG_INFO, _("Compiling starchart..."));

    parseMessages(inbox, *this);

    dat.postprocess(playing_set, data_set, playability);
    dis.postprocess(playing_set, data_set, playability);
    postprocessFleets(*this);
    syncTeamsFromAlliances(getAlliances(getRealPlayerId()));

    chart_bbox.initFromUniverse(dat);

    // inbox.sort(); // FIXME: configurable
    if (inbox.getCount() > 0) {
        GMessageFilter& filter = getMessageFilter();
        int n = 0;
        while (n < inbox.getCount() && filter.isFiltered(inbox.getHeading(n)))
            ++n;
        if (n < inbox.getCount())
            inbox.setCurrent(n);
    }

    console.write(LOG_INFO, _("=== Loading complete ==="));
    for (int i = 1; i <= NUM_PLAYERS; ++i)
        if (playing_set.contains(i))
            console.write(LOG_INFO, format(_("Playing player %d: %s"), i, player_racenames.getFullName(i)));
}

/** Recompute planet orbit flags (ship rings). */
void
GGameTurn::recomputeOrbitFlags()
{
    dat.recomputeOrbitFlags();
    dis.recomputeOrbitFlags();
}

GGameTurn::Player::Player(int pid)
    : outbox(pid)
{ }

GGameTurn::Player::~Player()
{ }

/**************************** Global Functions ***************************/

Signal0 sig_turn_changed;

void
setDisplayedTurn(Ptr<GGameTurn> disp)
{
    GGameTurn* orig = display_turn_internal;
    if (disp)
        display_turn_internal = disp;
    else
        display_turn_internal = current_turn_internal;

    current_ship_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_played_ships);
    current_planet_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_played_planets);
    current_base_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_played_bases);
    current_ion_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_ion_storms);
    current_minefield_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_minefields);
    current_anyplanet_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_any_planets);
    current_anyship_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_any_ships);
    current_histship_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_history_ships);
    current_ufo_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_ufos);
    current_fleet_type.setObjectType(&display_turn_internal->getCurrentUniverse().ty_fleets);

    current_ship_selection.setObjectType(&current_ship_type);
    current_planet_selection.setObjectType(&current_planet_type);
    current_base_selection.setObjectType(&current_base_type);
    current_ion_selection.setObjectType(&current_ion_type);
    current_minefield_selection.setObjectType(&current_minefield_type);
    current_histship_selection.setObjectType(&current_histship_type);
    current_ufo_selection.setObjectType(&current_ufo_type);
    current_fleet_selection.setObjectType(&current_fleet_type);

    if (orig != display_turn_internal)
        sig_turn_changed.raise();
}

void
initTurn()
{
    current_turn_internal = new GGameTurn();
    /* setDisplayedTurn(0);    FIXME: cannot call this as the turn is not yet usable;
                               this causes setUniverse to explode. Get rid of the init() stuff someday! */

    // Reset all selections.
    // This must be done so that the following setDisplayedTurn() sees a change in setObjectType() and updates the selections,
    // so we don't carry and old turn's selection around for too long.
    current_ship_selection.setObjectType(0);
    current_planet_selection.setObjectType(0);
    current_base_selection.setObjectType(0);
    current_ion_selection.setObjectType(0);
    current_minefield_selection.setObjectType(0);
    current_histship_selection.setObjectType(0);
    current_ufo_selection.setObjectType(0);
    current_fleet_selection.setObjectType(0);
}

void
initGameGlobals(string_t gamedir, string_t rootdir)
{
    initDirectories(gamedir, rootdir);
    initTurn();
    host.clear();
}
