/**
  *  \file game/ship.h
  *  \brief Ships
  *
  *  This defines all ship-related classes.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SHIP_H
#define PCC_V2_GAME_SHIP_H

#include <vector>
#include "game/struct.h"
#include "game/obj.h"
#include "game/objtypes.h"
#include "game/spec.h"
#include "game/cargo.h"
#include "cpluslib/maybe.h"
#include "io/io.h"              // mn16_t etc.
#include "game/unitscore.h"
#include "game/playerset.h"
#include "game/config.h"

class GUniverse;
class GTurnfile;
class GPlanet;
class GHullFunctionList;
class GMessageInformation;

/** Ship. We essentially have three kinds of ship:
    - current ship, i.e. one seen this turn, and possibly played
    - guessed ship, i.e. not seen this turn but we guess it's there
    - history ship, i.e. not seen this turn */
class GShip : public GMapObject {
 public:
    /// Transporters.
    enum Transporter {
        t_Unload,
        t_Transfer
    };
    enum Kind {
        NoShip,             /**< We do not know anything about this ship. */
        CurrentShip,        /**< This is a current ship. It is visible, we have a SHIP.DAT entry for it. */
        CurrentTarget,      /**< This is a current ship. It is visible, we have a TARGET.DAT entry for it. */
        CurrentUnknown,     /**< This is a current ship. It is visible, but we don't have any data but its mass (non-visual contact). */
        GuessedShip,        /**< This is a guessed ship. We have no current data for it, just history, and we display that on the map. */
        HistoryShip         /**< This is an old ship. We have history data for it, and it's not visible this turn. */
    };

    // Construction
    explicit GShip(int id);
    explicit GShip(const GShip& other);
    ~GShip();

    // Load and Save
    void       addShipData(const TShip& data, GPlayerSet source);
    void       addTargetData(const TShipTarget& target, GPlayerSet source);
    void       addShipXYData(TShipXY xy, GPlayerSet source);
    void       addRCEntry(int flag);
    void       addShipHistoryData(const TDbShip& data);
    void       addShipTrackEntry(const TDbShipTrackEntry& data, int turn);
    void       addShipFunction(int function);
    void       addMessageInformation(const GMessageInformation& info, int msgNr);

    void       getShipData(TShip& data) const;
    int        getShipTrackHeader(TDbShipTrack& header) const;
    void       getShipTrackEntry(TDbShipTrackEntry& entry, int slot) const;
    void       getShipHistoryData(TDbShip& the_data) const;

    void       internalCheck();
    void       combinedCheck1(GUniverse& univ);

    // GMapObject interface
    string_t   getName(Name which) const;
    int        getId() const;
    int        getOwner() const;
    GPoint     getPos() const;

    // Status inquiry
    bool       isVisible() const;
    bool       isReliablyVisible(int forPlayer) const;
    GPlayerSet getShipSource() const;
    Kind       getShipKind() const;
    bool       hasAnyShipData() const;
    bool       hasFullShipData() const;

    // History accessors
    int        getHistoryTimestamp(TDbShipTimestamp which_one) const;
    const TDbShipTrackEntry& getHistoryEntry(int turn) const;
    int        getHistoryNewestTurn() const;
    int        getHistoryOldestTurn() const;

    // Type accessors
    mp16_t     getMass() const;
    mp16_t     getHullId() const;
    void       setHullId(mp16_t h);
    GHull&     getHull() const;

    // Location accessors
    int        getOrbitPlanetId() const;
    bool       isCloningAt(const GPlanet& planet) const;

    // Owner accessors
    int        getRealOwner() const;
    int        getRemoteControlFlag() const;

    // Course accessors
    GPoint     getWaypoint() const;
    void       setWaypoint(GPoint pt);
    mn16_t     getWaypointDX() const;
    mn16_t     getWaypointDY() const;
    mp16_t     getHeading() const;

    mp16_t     getWarp() const;
    void       setWarp(mp16_t warp);
    bool       isHyperdriving() const;

    // Equipment accessors
    mp16_t     getEngineType() const;
    void       setEngineType(mp16_t engine);
    mp16_t     getBeamType() const;
    void       setBeamType(mp16_t type);
    mp16_t     getNumBeams() const;
    void       setNumBeams(mp16_t count);
    mp16_t     getNumBays() const;
    void       setNumBays(mp16_t count);
    mp16_t     getTorpType() const;
    void       setTorpType(mp16_t type);
    mp16_t     getNumTorpLaunchers() const;
    void       setNumTorpLaunchers(mp16_t count);

    // Mission accessors
    void       setName(const string_t& str);
    mp16_t     getMission() const;
    void       setMission(mp16_t m, mp16_t i, mp16_t t);
    mp16_t     getTowId() const;
    mp16_t     getInterceptId() const;
    mp16_t     getPrimaryEnemy() const;
    void       setPrimaryEnemy(mp16_t pe);
    mp16_t     getDamage() const;
    void       setDamage(mp16_t damage);
    mp16_t     getCrew() const;
    void       setCrew(mp16_t crew);
    string_t   getFCode() const;
    void       setFCode(const string_t& str);
    bool       isFCodeKnown() const;

    // Cargo accessors
    mp16_t     getAmmoRaw() const;
    mn32_t     getAmmo() const;
    mp16_t     getCargoRaw(GCargoType type) const;
    mn32_t     getCargo(GCargoType type) const;
    void       changeCargoRaw(GCargoType type, int amount);
    mn32_t     getFreeCargo() const;

    // Transporter accesssors
    TShipTransfer& getTransporter(Transporter which_one) throw();
    const TShipTransfer& getTransporter(Transporter which_one) const throw();
    bool       isTransporterActive(Transporter which_one) const throw();
    bool       isTransporterCancellable(Transporter which_one) const throw();
    void       cancelTranporter(Transporter which_one) throw();
    bool       isTransporterInUse(Transporter which_one) const throw();
    void       setTransporterInUse(Transporter which_one, bool in_use) throw();

    // Fleet accessors
    void       setFleetNumber(int fno);
    int        getFleetNumber() const;
    void       setFleetName(string_t name);
    const string_t& getFleetName() const;
    bool       isFleetLeader() const;
    bool       isFleetMember() const;

    // Prediction tracking
    void       setPredictedPos(GPoint pt);
    GPoint     getPredictedPos() const;

    // Function accessors
    bool       canDoSpecial(int basic_function) const;
    void       enumShipSpecificSpecials(GHullFunctionList& list) const;
    bool       hasAnyShipSpecificSpecials() const;

    const std::vector<int>& getAssociatedMessages() const;

    // Public members:
    GUnitScores unit_scores;
    GCargoArbiter cargo_arbiter;

 private:
    // Copy protection
    GShip& operator=(const GShip&);

    // Internal structures
    struct MapInfo {
        // Always known:
        const int id;        /**< Ship Id. Constant. */

        // Optional:
        mp16_t scanned_mass; /**< Scanned mass. Known for CurrentTarget and CurrentUnknown. For CurrentShip, taken from ship_data. */
        std::vector<int16_t> special_functions;

        // Derived:
        int orbit;           /**< Planet we're orbiting. 0 if none. Known if position known. */
        int rc_flag;         /**< Remote control. 0 if not RCd, -1 if forbidden, otherwise real owner. */
        Kind kind : 8;       /**< Kind of ship. */

        // Fleet information:
        int fleet_nr;
        string_t fleet_name;

        // Prediction:
        GPoint predicted_pos;

        MapInfo(int id);
        ~MapInfo();
    };
    struct ShipInfo;

    // Data:
    ShipInfo* ship_info;
    MapInfo   map_info;
    std::vector<int> associated_messages;

    // Source flags: we track the source of SHIP, TARGET, and SHIPXY records.
    GPlayerSet ship_source;
    GPlayerSet target_source;
    GPlayerSet xy_source;

    // Runtime data
    bool transporter_in_use[2];

    // Data access:
    ShipInfo&    createShipInfo();
    const TShip& getDisplayedShip() const;
    void         adjustShipTrack(int turn);
};

/****************************** Mine Laying ******************************/

struct GMineMissionData {
    int   mine_id;            /**< Id of mine field to manipulate. 0 to accept any. */
    int   owner;              /**< Required owner. */
    bool  type;               /**< Type of minefield. */
    int   num_torps;          /**< Number of torpedoes. */
    int32 num_units;          /**< Number of mine units. */
    bool  used_mission;
    bool  used_fcode;
};

bool parseMineLayingMission(const GShip& sh, const GUniverse& univ, GMineMissionData& md);
bool parseMineScoopMission(const GShip& sh, const GUniverse& univ, GMineMissionData& md);

/******************************* Chunneling ******************************/

struct GChunnelData {
    int   target;
    int   failure;
    int   kind;

    /* Failure modes. Note that the sequence of these values is used
       in parseChunnelMission. */
    enum {
        chf_MateDamaged = 1,
        chf_MateMoving  = 2,
        chf_MateTowed   = 4,
        chf_MateFuel    = 8,
        chf_MateAny     = 15,

        chf_Damaged     = 16,
        chf_Moving      = 32,
        chf_Towed       = 64,
        chf_Fuel        = 128,
        chf_Training    = 256,
        chf_SelfAny     = 496,

        chf_Distance    = 512
    };

    /* Kinds */
    enum {
        chk_Self   = 1,
        chk_Others = 2
    };
};

bool parseChunnelMission(const GShip& sh, const GUniverse& univ, GChunnelData& cd);
string_t formatChunnelFailures(int failures);


/**************************** Turn Fuel Usage ****************************/

int getTurnFuelUsage(const GShip& sh, const ConfigStdOption& opt, int turns);
int getTurnFuelUsage(int hull, int player, const ConfigStdOption& opt, int turns);

/**************************** Inline Functions ***************************/

void setOptimumWarp(GUniverse& univ, int sid);
int computeOptimumWarp(GUniverse& univ, int sid, GPoint pos, GPoint waypoint);

#endif
