/**
  *  \file gfx/pixmap.cc
  *  \brief Pixmap Implementation
  *
  *  GfxPixmap class, and image loading routines.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstdlib>
#include <cstring>
#include <SDL_error.h>
#include "defs.h"               // HAVE_LIBSDL_IMAGE
#include "gfx/pixmap.h"
#include "util/except.h"
#include "io/io.h"
#include "gfx/palette.h"
#include "io/rle.h"

#ifdef HAVE_LIBSDL_IMAGE
#  include <SDL_image.h>
#endif

#include "gfx/gfx-modetr.h"
#include "util/nls.h"
#include "gfx/clip.h"
#include "cpluslib/assert.h"
#include "io/sdlstr.h"
#include "io/streams.h"

/** \class GfxPixmap
    \brief Pixmap

    This represents a pixmap. A pixmap is a normal SDL_Surface, but
    with additional routines for managing it (format conversion,
    blitting it, ...).

    Const correctness: many methods which do not modify a surface
    are not marked const. The reason is that we're digging deep in
    SDL interna, and thigns like getSystemSurface() or getPixel()
    cannot, with good conscience, be const. */

/** Create new Pixmap. Creates a new pixmap with the specified
    parameters. */
GfxPixmap::GfxPixmap(int w, int h,
                     int depth,
                     uint32 rmask, uint32 gmask, uint32 bmask, uint32 amask)
    : GfxSurface(SDL_CreateRGBSurface(SDL_SWSURFACE,
                                      w, h,
                                      depth,
                                      rmask, gmask, bmask, amask),
                 true)
{
    if (!sfc)
        throw GraphicsIOException(SDL_GetError());
}

/** Use existing Surface. This wraps an existing surface in this
    GfxPixmap object.

    FIXME: This constructor should not be used by user code. It is
    currently used for constructing pictures from files. */
GfxPixmap::GfxPixmap(SDL_Surface* sfc, bool aown)
    : GfxSurface(sfc, aown)
{ }

/** Create pixmap in same format as canvas. The pixmap will have the
    same pixel format as the given canvas, simplifying copying of
    pixels and blitting. */
GfxPixmap::GfxPixmap(int w, int h, GfxSurface& copy_format)
    : GfxSurface(w, h, copy_format.getSystemSurface()->format)
{
    SDL_PixelFormat* format = copy_format.getSystemSurface()->format;
    if (format->BitsPerPixel <= 8)
        SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL,
                       format->palette->colors, 0, format->palette->ncolors);
}

/** Convert to specified format.
    \returns new surface, converted to specified format; 0 on error. */
Ptr<GfxPixmap>
GfxPixmap::convertTo(GfxSystemPixelFormat fmt) const
{
    // RETARGET: assumes GfxSystemPixelFormat is SDL_PixelFormat*
    SDL_Surface* sfc = SDL_ConvertSurface(this->sfc, fmt, SDL_SWSURFACE);
    if (sfc)
        return new GfxPixmap(sfc, true);
    else
        return 0;
}

/** Convert to screen format.
    \returns new surface, converted to pixel format of screen; 0 on error. */
Ptr<GfxPixmap>
GfxPixmap::convertToScreenFormat() const
{
    SDL_Surface* sfc = SDL_DisplayFormat(this->sfc);
    if (sfc) {
        return new GfxPixmap(sfc, true);
    } else {
        return 0;
    }
}

/** Blit Pixmap on Canvas. Displays this pixmap on the specified
    canvas, at the specified position. */
void
GfxPixmap::blit(GfxCanvas& can, int x, int y)
{
    can.blitSurface(GfxPoint(x, y), *this, GfxRect(getSize()));
}

/** Blit Pixmap on Canvas. Displays a sub-area of this pixmap on
    the specified canvas.
    \param can  Canvas
    \param x,y  Coordinates on canvas
    \param subr Area of this pixmap to use, in pixmap coordinates.

    \note The /x/ and /y/ parameters specify the coordinates of the
    upper-left point after clipping the pixmap at /subr/. For
    GfxCanvas::blitSurface(), they specify the coordinates before
    clipping. */
void
GfxPixmap::blit(GfxCanvas& can, int x, int y, GfxRect subr)
{
    x -= subr.x;
    y -= subr.y;
    subr.clipSize(sfc->w, sfc->h);
    can.blitSurface(GfxPoint(x, y), *this, subr);
}

/** Blit Pixmap on canvas. Ensures that the area x+wi/y+he is covered,
    filling with color 0. */
void
GfxPixmap::blitSized(GfxCanvas& can, int x, int y, int wi, int he)
{
    GfxRect pix_rect(getSize());

    if (pix_rect.h > he) {
        /* pixmap is bigger than our width -- reduce */
        pix_rect.y += (pix_rect.h - he) / 2;
        pix_rect.h = he;
    } else if (pix_rect.h < he) {
        /* pixmap is too small. Draw outside */
        int overlap = he - pix_rect.h, otop = overlap/2, obot = overlap - otop;
        can.drawBar(GfxRect(x, y,             wi, otop), 0, 0, SOLID_FILL, GFX_ALPHA_OPAQUE);
        can.drawBar(GfxRect(x, y + he - obot, wi, obot), 0, 0, SOLID_FILL, GFX_ALPHA_OPAQUE);
        y += otop;
        he = pix_rect.h;
    }

    if (pix_rect.w > wi) {
        /* pixmap is bigger than our width -- reduce */
        pix_rect.x += (pix_rect.w - wi) / 2;
        pix_rect.w = wi;
    } else if (pix_rect.w < wi) {
        /* pixmap is too small. Draw outside */
        /* |   |XXXX|   |
           x            x+wi */
        int overlap = wi - pix_rect.w, oleft = overlap/2, oright = overlap - oleft;
        can.drawBar(GfxRect(x,               y, oleft,  he), 0, 0, SOLID_FILL, GFX_ALPHA_OPAQUE);
        can.drawBar(GfxRect(x + wi - oright, y, oright, he), 0, 0, SOLID_FILL, GFX_ALPHA_OPAQUE);
        x += oleft;
        wi = pix_rect.w;
    }
    ASSERT(pix_rect.w == wi && pix_rect.h == he);
    can.blitSurface(GfxPoint(x - pix_rect.x, y - pix_rect.y), *this, pix_rect);
}

/** Flip Pixmap Vertically. Flips the pixmap data so that the pixmap
    is now upside-down. */
void
GfxPixmap::flipVertical()
{
    if (!sfc->h)
        return;

    // FIXME: this operation can be done without temporary storage
    unsigned ptr1 = 0;
    unsigned ptr2 = sfc->h - 1;
    Array<char> tmp(sfc->pitch);

    while (ptr1 < ptr2) {
        char* chr1 = static_cast<char*>(sfc->pixels) + sfc->pitch * ptr1;
        char* chr2 = static_cast<char*>(sfc->pixels) + sfc->pitch * ptr2;
        char* tmpp = tmp;

        std::memcpy (tmpp, chr1, sfc->pitch);
        std::memcpy (chr1, chr2, sfc->pitch);
        std::memcpy (chr2, tmpp, sfc->pitch);

        ++ptr1;
        --ptr2;
    }
}

template<class T>
void
flipHorizontalLine(uint8* data, unsigned width)
{
    unsigned l = 0;
    unsigned r = width-1;
    while (l < r) {
        uint8* pl = T::add (data, l);
        uint8* pr = T::add (data, r);
        typename T::pix_type tmp = T::peek (pl);
        T::poke (pl, T::peek (pr));
        T::poke (pr, tmp);
        ++l;
        --r;
    }
}

/** Flip Pixmap Horizontally. Flips the pixmap so that left becomes
    right and vice versa. */
void
GfxPixmap::flipHorizontal()
{
    uint8* pix = static_cast<uint8*>(sfc->pixels);

    if (!sfc->w)
        return;

    for (int i = 0; i < sfc->h; ++i) {
        GFX_MODE_SWITCH(flipHorizontalLine, (pix, sfc->w));
        pix += sfc->pitch;
    }
}

/** Clone this pixmap. Returns an exact copy. */
Ptr<GfxPixmap>
GfxPixmap::clone() const
{
    Ptr<GfxPixmap> pix = convertTo(sfc->format);
    pix->copySettingsFrom(*this);
    return pix;
}

/** Tile area with pixmap.
    \param can    Target canvas
    \param area   Area to tile
    \param alt    Alteration of X coordinate
    \see tileAnchored() */
void
GfxPixmap::tile(GfxCanvas& can, const GfxRect& area, int alt)
{
    tileAnchored(can, area, area.x, area.y, alt);
}

/** Tile area with pixmap, anchored.
    \param can    Target canvas
    \param area   Area to tile
    \param ax,ay  Anchor point
    \param alt    Alteration of X coordinate

    To have multiple tiled regions fit to each other nicely, use the same
    anchor point. Using tile() instead will anchor the tiles at the top-left
    point of the area; this way, the join points will be visible.

    The alteration specifies how the X anchor point is alterated on each
    line. With alt=0, you'll get a "chequered" pattern; alt=16 starts the
    first line at ax, the second at ax-16, the third at ax again, etc. This
    makes large tiled regions look much nicer. */
void
GfxPixmap::tileAnchored(GfxCanvas& can, const GfxRect& area, int ax, int ay, int alt)
{
    int y2 = area.y + area.h, x2 = area.x + area.w;
    int y = ay, x0 = ax;
    int mult = -1;

    while (y < y2) {
        int x = x0;
        while (x < x2) {
            GfxRect blit_rect(getSize());

            GfxRect rclip = area;
            rclip.moveBy(-x, -y);
            blit_rect.clipRect(rclip);
            if (blit_rect.exists())
                can.blitSurface(GfxPoint(x, y), *this, blit_rect);

            x += sfc->w;
        }
        x0 += mult * alt;
        mult = -mult;
        y += sfc->h;
    }
}

/** Get current color-key. Color-keyed pixels are transparent when this
    bitmap is blitted somewhere.
    \return GFX_TRANSPARENT if there is no color-key active for this
    pixmap; pixel value otherwise.
    \see isColorKeyed() */
uint32
GfxPixmap::getColorKey() const
{
    if (sfc->flags & SDL_SRCCOLORKEY)
        return sfc->format->colorkey;
    else
        return GFX_TRANSPARENT;
}

/** Set color-key. Color-keyed pixels are transparent when this
    bitmap is blitted somewhere.
    \param color GFX_TRANSPARENT to disable color-keying,
    otherwise a color */
void
GfxPixmap::setColorKey(uint32 color)
{
    if (color == GFX_TRANSPARENT)
        SDL_SetColorKey(sfc, 0, 0);
    else
        SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, color);
}

/** Get per-surface alpha. The per-surface alpha value is used to
    blend the complete surface when blitting it somewhere. This
    value is only meaningful when isBlitAlphaSupported() is true. */
uint8
GfxPixmap::getBlitAlpha() const
{
    if (sfc->flags & SDL_SRCALPHA)
        return sfc->format->alpha;
    else
        return SDL_ALPHA_OPAQUE;
}

/** Set per-surface alpha. The per-surface alpha value is used to
    blend the complete surface when blitting it somewhere. If
    isBlitAlphaSupported() is false, this call is ignored. */
void
GfxPixmap::setBlitAlpha(uint8 alpha)
{
    if (alpha == SDL_ALPHA_OPAQUE)
        SDL_SetAlpha(sfc, 0, 0);
    else
        SDL_SetAlpha(sfc, SDL_SRCALPHA, alpha);
}

/** Check whether alpha-blit is supported. If alpha-blit is not
    supported, the surface is an RGBA surface and individual
    pixels must be alpha-blended. */
bool
GfxPixmap::isBlitAlphaSupported() const
{
    return (sfc->format->Amask == 0);
}

/** Copy blit settings from other pixmap. Used during
    initialisation. */
void
GfxPixmap::copySettingsFrom(const GfxPixmap& other)
{
    setColorKey(other.getColorKey());
    if (isBlitAlphaSupported())
        setBlitAlpha(other.getBlitAlpha());
}

/** Compute bounding rectangle. The bounding rectangle is the smallest
    rectangle such that everything in this pixmap outside that
    rectangle has color /color/. */
GfxRect
GfxPixmap::computeBoundingRectangle(const uint32 color)
{
    GfxRect res(getSize());

    // locate bottom
    while (res.h > 0) {
        for (unsigned int x = 0; x < res.w; ++x)
            if (getPixel(GfxPoint(x, res.h-1)) != color)
                goto bottom_found;
        --res.h;
    }
 bottom_found:

    // empty?
    if (res.h == 0)
        return GfxRect();

    // locate right side
    while (res.w > 0) {
        for (unsigned int y = 0; y < res.h; ++y)
            if (getPixel(GfxPoint(res.w-1, y)) != color)
                goto right_found;
        --res.w;
    }
 right_found:

    // locate top
    while (res.h > 0) {
        for (unsigned int x = 0; x < res.w; ++x)
            if (getPixel(GfxPoint(x, res.y)) != color)
                goto top_found;
        --res.h, ++res.y;
    }
 top_found:

    // locate left side
    while (res.w > 0) {
        for (unsigned int y = 0; y < res.h; ++y)
            if (getPixel(GfxPoint(res.x, res.y + y)) != color)
                goto left_found;
        --res.w, ++res.x;
    }
 left_found:

    return res;
}

/** Crop this pixmap. This first computes the bounding rectangle, and
    then creates a new pixmap which only contains the "interesting"
    contents of this one.
    \return new pixmap, null if bounding rectangle is empty
    \see computeBoundingRectangle() */
Ptr<GfxPixmap>
GfxPixmap::cloneCropped(const uint32 color)
{
    GfxRect r = computeBoundingRectangle(color);
    if (!r.exists())
        return 0;

    Ptr<GfxPixmap> pix = new GfxPixmap(r.w, r.h, *this);
    pix->copySettingsFrom(*this); // FIXME: position?
    blit(*pix, 0, 0, r);
    return pix;
}

/***************************** Image Loading *****************************/

/** Convert PCC v1 color to equivalent PCC v2 color. */
static inline unsigned char
convertColor(unsigned char c)
{
    if (c >= 0x30)
        return c - 0x20;
    else
        return c & 0x0F;
}

static inline unsigned char
convertColor16(unsigned char c)
{
    if (c >= 10 && c < 15)
        return c-10 + 160;
    else
        return c;
}

/** Load .cc image. Those are used by PCC v1 in 16 color mode.
    \pre magic number has been read.
    \assumes stream need not be seekable.

    File format:
    - 2 bytes magic ('CC')
    - 2 bytes width
    - 2 bytes height
    - 5x3 bytes palette (slots 10..14, entry ranges 0..63)
    - pixel data */
static Ptr<GfxPixmap>
loadCCImage(Stream& s)
{
    // FIXME: should put "user" colors elsewhere to allow reordering
    // in 8-bit mode
    char head[4 + 5*3];
    if (s.read(head, sizeof head) != sizeof head)
        return 0;

    unsigned w = getUint16(head, 0);
    unsigned h = getUint16(head, 2);
    if (w == 0 || h == 0 || w > 1000 || h > 1000)
        return 0;

    unsigned bytes_per_line = (w+1) / 2;
    Array<char> line (bytes_per_line);

    /* Image seems to be valid. Load it. */
    // FIXME: this digs into SDL_Surface intrinsics
    Ptr<GfxPixmap> pix = new GfxPixmap(w, h, 8, 0, 0, 0, 0);
    SDL_Surface* sfc = pix->getSystemSurface();

    ASSERT(!SDL_MUSTLOCK(sfc));

    for (unsigned i = 0; i < h; ++i) {
        char* pixel_data = static_cast<char*>(sfc->pixels) + sfc->pitch * i;
        char* pixel_src = line;

        s.readT(line, bytes_per_line);
        for (unsigned x = 0; x < w; ++x)
            if (x & 1)
                *pixel_data++ = convertColor16((*pixel_src++ >> 4) & 0x0F);
            else
                *pixel_data++ = convertColor16(*pixel_src & 0x0F);
    }

    /* set up palette */
    SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, standard_palette, 0, 160);
    for (int i = 0; i < 5; ++i) {
        SDL_Color c = {
            head[4 + 3*i] * 255 / 63,
            head[5 + 3*i] * 255 / 63,
            head[6 + 3*i] * 255 / 63,
            0
        };
        SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, &c, 160+i, 1);
    }
    return pix;
}

/** Load .cc image. Those are used by PCC v1 in 256 color mode.
    \pre magic number has been read.
    \assumes stream need not be seekable.

    File format:
    - 2 bytes magic ('CD')
    - 2 bytes width
    - 2 bytes height
    - 64x3 bytes palette (slots 192..255, ranges 0..63)
    - pixel data

    Palette will be shuffled accordingly. */
static Ptr<GfxPixmap>
loadCDImage(Stream& s)
{
    char head[4 + 64*3];
    if (s.read(head, sizeof head) != sizeof head)
        return 0;

    unsigned w = getUint16(head, 0);
    unsigned h = getUint16(head, 2);
    if (w == 0 || h == 0 || w > 1000 || h > 1000)
        return 0;

    /* Image seems to be valid. Load it. */
    Ptr<GfxPixmap> pix = new GfxPixmap(w, h, 8, 0, 0, 0, 0);
    SDL_Surface* sfc = pix->getSystemSurface();

    ASSERT(!SDL_MUSTLOCK(sfc));

    for (unsigned i = 0; i < h; ++i) {
        char* pixel_data = static_cast<char*>(sfc->pixels) + sfc->pitch * i;

        s.readT(pixel_data, w);
        for (unsigned x = 0; x < w; ++x)
            pixel_data[x] = convertColor(pixel_data[x]);
    }

    /* set up palette */
    SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, standard_palette, 0, 160);
    for (int i = 0; i < 64; ++i) {
        SDL_Color c = {
            head[4 + 3*i] * 255 / 63,
            head[5 + 3*i] * 255 / 63,
            head[6 + 3*i] * 255 / 63,
            0
        };
        SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, &c, 160+i, 1);
    }
    return pix;
}

/** Load .gfx image. These are used for various things in PCC v1, and are
    the native format of the graphics library PCC v1 is based on. Unlike
    the other formats, we do not translate those; we assume they have the
    right palette.

    File format:
    - 2 bytes magic (0, 8)
    - 2 bytes width
    - 2 bytes height
    - pixel data. FF means transparent */
static Ptr<GfxPixmap>
loadGFXImage(Stream& s)
{
    char head[4];
    if (s.read(head, 4) != 4)
        return 0;

    unsigned w = getUint16(head, 0);
    unsigned h = getUint16(head, 2);
    if (w == 0 || h == 0 || w > 1000 || h > 1000)
        return 0;

    /* Image seems to be valid. Load it. */
    Ptr<GfxPixmap> pix = new GfxPixmap(w, h, 8, 0, 0, 0, 0);
    SDL_Surface* sfc = pix->getSystemSurface();

    pix->lock();

    bool saw_ff = false;
    for (unsigned i = 0; i < h; ++i) {
        char* pixel_data = static_cast<char*>(sfc->pixels) + sfc->pitch * i;
        s.readT(pixel_data, w);

        if (!saw_ff)
            for (unsigned j = 0; j < w; ++j)
                if (static_cast<unsigned char>(pixel_data[j]) == 0xFFU) {
                    saw_ff = true;
                    break;
                }
    }
    pix->unlock();
    pix->update(GfxRect(pix->getSize()));

    /* set up palette */
    SDL_SetPalette(sfc, SDL_LOGPAL | SDL_PHYSPAL, standard_palette, 0, NUM_STANDARD_COLORS);
    if (saw_ff)
        SDL_SetColorKey(sfc, SDL_SRCCOLORKEY, 0xFFU);

    return pix;
}

/** Load image. Attempts to automatically figure out the type.
    If was_compr is true, s is a decompressed RLE stream; these
    may contain only .cc/.cd/.gfx images as used in PCC v1.
    If was_compr is false, stream must be seekable and can contain
    anything SDL (SDL_image) can read. */
static Ptr<GfxPixmap>
loadImageInternal(Stream& s, bool was_compr)
{
    char magic[2];

    s.seek(0);
    if (s.read(magic, 2) == 2) {
        if (magic[0] == 'C' && magic[1] == 'C')
            return loadCCImage(s);
        else if (magic[0] == 'C' && magic[1] == 'D')
            return loadCDImage(s);
        else if (magic[0] == 0 && magic[1] == 8)
            return loadGFXImage(s);
    }

    if (!was_compr) {
        SDL_Surface* sfc;
        s.seek(0);
        StreamSDLInterface iface(s);
#  ifdef HAVE_LIBSDL_IMAGE
        sfc = IMG_Load_RW(&iface, 0);
#  else
        sfc = SDL_LoadBMP_RW(&iface, 0);
#  endif
        if (sfc)
            return new GfxPixmap(sfc, true);
    }
    return 0;
}

/** Image Loading, main entry point. Load image contained in stream s.
    \pre stream is at position 0 and seekable. */
Ptr<GfxPixmap>
loadImage(Stream& s)
{
    Ptr<GfxPixmap> pix = loadImageInternal(s, false);
    if (pix.ptr())
        return pix;

    s.seek(0);
    RleReaderStream r(s);
    return loadImageInternal(r, true);
}
