#!/usr/bin/perl -w
#
#  Convert PHost help file into PCC2 help file
#
#  Quick & Dirty.
#
use strict;
use SimpleXML;
use Data::Dumper;

# Configuration
my $dir = "../phost/doc4";       # Path to PHost XML files
my $macros = "macros.txt";       # Name of macro file
my $prefix = "phost4-";          # Prefix for generated pages
my $extra;
foreach (@ARGV) {
    if (/^--?dir=(.*)/) {
        $dir = $1;
    } elsif (/^--?macros=(.*)/) {
        $macros = $1;
    } elsif (/^--?prefix=(.*)/) {
        $prefix = $1;
    } elsif (/^--?extra=(.*)/) {
        $extra = $1;
    } else {
        die "Unknown option: '$_'";
    }
}

# Load macros
my %macros = ('$' => '$');
open MAC, "<", "$dir/$macros" or die "$dir/$macros: $!";
while (<MAC>) {
    if (/&\s*set\s+(\S+)\s*(.*)/) {
        $macros{$1} = $2;
    }
}
close MAC;

my %entities = (nbsp => "&#160;", 
                lt => "&#60;",
                le => "&#8804;",
                gt => "&#62;",
                ge => "&#8805;",
                quot => "&#34;",
                amp => "&#38;",
                regonly => "Registered Players Only");

# Load XML
my %files;
opendir DIR, $dir or die "$dir: $!";
while (defined(my $name = readdir(DIR))) {
    if ($name =~ /^(.*)\.xml$/) {
        my $base = $1;
        open FILE, "<", "$dir/$name" or die "$dir/$name: $!";
        my $text = join('', <FILE>);
        close FILE;

        # Preprocessing
        $text =~ s/\$([\w\d]+|.)/exists $macros{$1} ? $macros{$1} : ((print STDERR "Unknown macro: '$1'\n"), "")/eg;
        $text =~ s/&([\w\d]+);/exists $entities{$1} ? $entities{$1} : "&$1;"/eg;
        if ($text =~ /encoding=\"iso-8859-1\"/) {
            $text =~ s/([\x80-\xFF])/chr(0xC0 + (ord($1)>>6)).chr(0x80 + (ord($1) & 63))/eg;
        }

        lex_set_text($text);
        lex();
        $files{$base} = parse_document();
    }
}

# Preprocessing: Links
my %link_titles;     # Title for each link (full page id -> name)
my %link_targets;    # Alternate target for each link (full page id -> full page id), used to redirect inner <config> to outer <config> page
my %page_index;      # Indexes (xml basename -> key -> full page id)
foreach (sort keys %files) {
    preprocess_links($_, $files{$_});
}

# Preprocessing: Extract elements
my %elems;           # Saved fragments
foreach (sort keys %files) {
    preprocess_fragments($files{$_});
}

# Produce output
print "<?xml version=\"1.0\" ?>\n";
print "<help>\n";
print " <page id=\"toc\">\n";
print "  <h1>PHost 4 Help</h1>\n";
print "  <ul><li><a href=\"${prefix}index\">Index</a></li></ul>\n";
print " </page>\n";
foreach (sort keys %files) {
    render_pages($_, $files{$_});
}
if (defined($extra)) {
    open EXTRA, '<', $extra or die "$extra: $!";
    print <EXTRA>;
    close EXTRA;
}
print "</help>\n";


############################## Preprocessing ##############################

sub preprocess_links {
    my ($label, $pp) = @_;
    foreach my $p (@$pp) {
        if (is_tag($p, 'document')) {
            my $id = "$prefix$label";
            $link_titles{$id} = page_name($p);
            preprocess_links($label, $p->[2]);
        } elsif (is_tag($p, 'chapter', 'section', 'subsection', 'mission') && exists $p->[1]{anchor}) {
            my $id = "$prefix$p->[1]{anchor}";
            $link_titles{$id} = page_name($p);
            preprocess_links($label, $p->[2]);
            if (is_tag($p, 'mission')) {
                $page_index{$label}{page_name($p)} = $id;
            }
        } elsif (is_tag($p, 'config')) {
            my $id = $prefix."cf-".$p->[1]{name};
            $link_titles{$id} = $p->[1]{name};
            $page_index{$label}{$p->[1]{name}} = $id;
            my @ch = @{$p->[2]};
            while (@ch) {
                if (is_tag($ch[0], 'config')) {
                    my $child_id = $prefix."cf-".$ch[0][1]{name};
                    $link_titles{$child_id} = $ch[0][1]{name};
                    $link_targets{$child_id} = $id;
                    $page_index{$label}{$ch[0][1]{name}} = $id;
                    shift @ch;
                } elsif ($ch[0] !~ /\S/) {
                    shift @ch;
                } else {
                    last;
                }
            }
        } elsif (ref($p)) {
            preprocess_links($label, $p->[2]);
        }
    }
}

sub preprocess_fragments {
    my $pp = shift;
    foreach my $p (@$pp) {
        if (ref($p)) {
            if (exists $p->[1]{saveid}) {
                $elems{$p->[1]{saveid}} = $p;
            }
            preprocess_fragments($p->[2]);
        }
    }
}


################################ Rendering ################################

sub render_pages {
    my ($label, $pp) = @_;
    foreach my $p (@$pp) {
        if (is_tag($p, 'document')) {
            print " <page id=\"$prefix$label\">\n";
            print "  <h1>", page_name($p), "</h1>\n";
            my @content = @{$p->[2]};
            shift @content;
            render_content($label, \@content);
            print " </page>\n";
            render_pages($label, $p->[2]);
        } elsif (is_tag($p, 'chapter', 'section', 'subsection', 'mission') && exists $p->[1]{anchor}) {
            print " <page id=\"$prefix$p->[1]{anchor}\">\n";
            print "  <h1>", $p->[1]{name}, "</h1>\n";
            render_content($label, $p->[2]);
            print " </page>\n";
            render_pages($label, $p->[2]);
        } elsif (is_tag($p, 'config')) {
            print " <page id=\"$prefix", "cf-", "$p->[1]{name}\">\n";
            render_option($p);
            my @ch = @{$p->[2]};
            while (@ch) {
                if (is_tag($ch[0], 'config')) {
                    render_option($ch[0]);
                    shift @ch;
                } elsif ($ch[0] !~ /\S/) {
                    shift @ch;
                } else {
                    last;
                }
            }
            render_content($label, \@ch);
            print " </page>\n";
        } elsif (is_tag($p, 'index')) {
            print " <page id=\"$prefix$label-ALPHA\">\n";
            print "  <h1>$macros{ph_alpha}</h1>\n";
            my @keys = sort {uc($a) cmp uc($b)} keys %{$page_index{$label}};
            if (@keys) {
                my $last_first_letter = '';
                foreach my $k (@keys) {
                    my $first_letter = uc(substr($k, 0, 1));
                    if ($first_letter ne $last_first_letter) {
                        print "  </ul>\n" if $last_first_letter ne '';
                        print "  <h2>$first_letter</h2>\n";
                        print "  <ul class=\"compact\">\n";
                        $last_first_letter = $first_letter;
                    }
                    print "   <li><a href=\"$page_index{$label}{$k}\">$k</a></li>\n";
                }
                print "  </ul>\n" if $last_first_letter ne '';
            } else {
                print STDERR "Empty index for '$label'\n";
            }
            print " </page>\n";
        } elsif (ref($p)) {
            render_pages($label, $p->[2]);
        }
    }
}

sub render_option {
    my $p = shift;
    print "  <h1>", $p->[1]{name}, "</h1>\n";
    print "  <dl>\n";
    if (exists $p->[1]{range}) {
        print "  <di term=\"Range\">", $p->[1]{range}, "</di>\n";
    }
    if (exists $p->[1]{enum}) {
        print "  <di term=\"Values\">", $p->[1]{enum}, "</di>\n";
    }
    if (exists $p->[1]{default}) {
        print "  <di term=\"Default Value\">", $p->[1]{default}, "</di>\n";
    }
    if (exists $p->[1]{incompatible}) {
        print "  <di term=\"Incompatible!\"></di>\n";
    }
    if (exists $p->[1]{arrayize}) {
        print "  <di term=\"Arrayized!\"></di>\n";
    }
    if (exists $p->[1]{array}) {
        print "  <di term=\"Array\"></di>\n";
    }
    if (exists $p->[1]{phost}) {
        print "  <di term=\"PHost only!\"></di>\n";
    }
    if (exists $p->[1]{optional}) {
        print "  <di term=\"Optional!\"></di>\n";
    }
    print "  </dl>\n";
}

sub render_content {
    my ($label, $pp) = @_;
    foreach my $p (@$pp) {
        if (ref($p)) {
            if (is_tag($p, 'p')) {
                print "<p>";
                if (exists $p->[1]{prefix}) {
                    print "<b>", $p->[1]{prefix}, ": </b>";
                }
                render_content($label, $p->[2]);
                print "</p>";
            } elsif (is_tag($p, 'a')) {
                my $t = map_href($p->[1]{href});
                if (defined($t)) {
                    print "<a href=\"$t\">";
                    if (@{$p->[2]}) {
                        render_content($label, $p->[2]);
                    } elsif (exists $link_titles{$t}) {
                        print $link_titles{$t};
                    } else {
                        print STDERR "Link without text: $p->[1]{href}\n";
                    }
                    print "</a>";
                } else {
                    render_content($label, $p->[2]);
                }
            } elsif (is_tag($p, 'cfg')) {
                my $key = $p->[2][0];
                if (defined($key) && !ref($key)) {
                    print "<a href=\"", $prefix, "cf-", $key, "\"><tt>$key</tt></a>";
                }
            } elsif (is_tag($p, 'attn')) {
                print "<font color=\"red\">&#9658; Note: </font>";
            } elsif (is_tag($p, 'file')) {
                print "<tt>";
                render_content($label, $p->[2]);
                print "</tt>";
            } elsif (is_tag($p, 'arg')) {
                print "<p><b>", $p->[1]{name}, ": </b>";
                render_content($label, $p->[2]);
                print "</p>";
            } elsif (is_tag($p, 'new')) {
                print "<p><font color=\"green\">(", $p->[1]{version}, ") </font>";
                render_content($label, $p->[2]);
                print "</p>";
            } elsif (is_tag($p, 'em', 'b', 'ol', 'ul', 'li', 'tt', 'br', 'tr', 'td', 'tn', 'th', 'table')) {
                my $tag = $p->[0];
                print "<$tag>";
                render_content($label, $p->[2]);
                print "</$tag>";
            } elsif (is_tag($p, 'gr')) {
                print "<font color=\"green\">";
                render_content($label, $p->[2]);
                print "</font>";
            } elsif (is_tag($p, 'ex')) {
                print "<pre>";
                render_content($label, $p->[2]);
                print "</pre>";
            } elsif (is_tag($p, 'diff')) {
                print "<p>";
                render_content($label, $p->[2]);
                print "</p>";
            } elsif (is_tag($p, 'dl')) {
                # PCC2 cannot render dl/dt/dd in the way HTML renders it, so map it to <ul>
                print "<ul>";
                render_content($label, $p->[2]);
                print "</ul>";
            } elsif (is_tag($p, 'dt')) {
                print "<li>";
                render_content($label, $p->[2]);
                print "</li>";
            } elsif (is_tag($p, 'dd')) {
                print "<li>&#9658;&#160;";
                render_content($label, $p->[2]);
                print "</li>";
            } elsif (is_tag($p, 'text', 'footnote')) {
                render_content($label, $p->[2]);
            } elsif (is_tag($p, 'elem')) {
                my $eid = $p->[1]{id};
                if (exists $elems{$eid}) {
                    render_content($label, $elems{$eid}[2]);
                } else {
                    print STDERR "Saved element '$eid' not found\n";
                }
            } elsif (is_tag($p, 'section', 'subsection', 'mission', 'chapter')) {
                if (exists $p->[1]{anchor}) {
                    print "<p><a href=\"$prefix$p->[1]{anchor}\">$p->[1]{name}</a></p>\n";
                } else {
                    print "<h2>$p->[1]{name}</h2>\n";
                    render_content($label, $p->[2]);
                }
            } elsif (is_tag($p, 'index')) {
                print "<p><a href=\"$prefix$label-ALPHA\">$macros{ph_alpha}</a></p>\n";
            } elsif (is_tag($p, 'config')) {
                print "<p><a href=\"$prefix", "cf-", "$p->[1]{name}\">$p->[1]{name}</a></p>\n";
            } else {
                print STDERR "Unknown tag: $p->[0]\n";
            }
        } else {
            print $p;
        }
    }
}

################################ Utilities ################################

sub is_tag {
    my $p = shift;
    return ref($p) && grep {$p->[0] eq $_} @_;
}

sub page_name {
    my $p = shift;
    if (is_tag($p, 'chapter', 'section', 'subsection', 'mission')) {
        return $p->[1]{name};
    } elsif (is_tag($p, 'document')) {
        if (ref($p->[2][0]) && $p->[2][0][0] eq 'title') {
            return $p->[2][0][2][0];
        } else {
            return '';
        }
    } else {
        return '';
    }
}

sub map_href {
    my $x = shift;
    if (defined($x)) {
        if ($x =~ /^[*#](.*)/) {
            my $id = $prefix.$1;
            if (exists $link_targets{$id}) {
                return $link_targets{$id};
            } else {
                return $id;
            }
        } else {
            # Whitelist of links we willfully ignore
            print STDERR "Link not mapped: '$x'\n"
                unless $x =~ /^(http|prog|phost|sr|news|vgb|pdk|phostdev|plist|vgap)(:|$)/;
            return undef;
        }
    } else {
        return undef;
    }
}
