/**
  *  \file u/t_util_misc.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "util/misc.h"
#include "u/t_util.h"

#include <vector>

void
UtilMiscTestSuite::testStringMatch()
{
    TS_ASSERT(stringMatch("ENglish", "english"));
    TS_ASSERT(stringMatch("ENglish", "en"));
    TS_ASSERT(stringMatch("ENglish", "eng"));
    TS_ASSERT(stringMatch("ENglish", "ENGLISH"));
    TS_ASSERT(!stringMatch("ENglish", "e"));

    TS_ASSERT(stringMatch("ENGLISH", "english"));
    TS_ASSERT(!stringMatch("ENGLISH", "englis"));
    TS_ASSERT(!stringMatch("ENGLISH", "en"));
}

void
UtilMiscTestSuite::testStringValue()
{
    string_t a = "126";
    string_t b = "42km";
    string_t c = "45 ";
    string_t d = " 17 ";
    string_t e = "   99  miles ";
    int res;
    string_t::size_type pos;

    TS_ASSERT(stringValue(a, res));
    TS_ASSERT_EQUALS(res, 126);

    TS_ASSERT(!stringValue(b, res));
    TS_ASSERT(!stringValue(b, res, pos));
    TS_ASSERT_EQUALS(pos, 2U);
    TS_ASSERT(stringValue(b.substr(0, pos), res));
    TS_ASSERT_EQUALS(res, 42);

    TS_ASSERT(stringValue(c, res));
    TS_ASSERT_EQUALS(res, 45);

    TS_ASSERT(stringValue(d, res));
    TS_ASSERT_EQUALS(res, 17);

    TS_ASSERT(!stringValue(e, res));
    TS_ASSERT(!stringValue(e, res, pos));
    TS_ASSERT_EQUALS(pos, 7U);
    TS_ASSERT(stringValue(e.substr(0, pos), res));
    TS_ASSERT_EQUALS(res, 99);
}

void
UtilMiscTestSuite::testParseFlags()
{
    int result;

    /*
       regular flags only
     */
    // very basic invocation
    result = 99;
    TS_ASSERT(parseFlags("ace", "XABCDE", &result, 0, false));
    TS_ASSERT_EQUALS(result, 42);

    // we support caps as well
    result = 99;
    TS_ASSERT(parseFlags("ACE", "XABCDE", &result, 0, false));
    TS_ASSERT_EQUALS(result, 42);

    // correct bit alignment (first character of pattern is bit 0)
    result = 99;
    TS_ASSERT(parseFlags("xace", "XABCDE", &result, 0, false));
    TS_ASSERT_EQUALS(result, 43);

    // refuse mismatching input
    TS_ASSERT(!parseFlags("back", "XABCDE", &result, 0, false));

    // accept mismatching input if allowed
    result = 99;
    TS_ASSERT(parseFlags("back", "XABCDE", &result, 0, true));
    TS_ASSERT_EQUALS(result, 14);

    // refuse input looking like race list if not configured
    TS_ASSERT(!parseFlags("ace-1", "XABCDE", &result, 0, false));

    // accept input looking like race list if mismatch allowed
    result = 99;
    TS_ASSERT(parseFlags("ace-1", "XABCDE", &result, 0, true));
    TS_ASSERT_EQUALS(result, 42);

    // correctly parse input looking like race list if mismatch allowed
    result = 99;
    TS_ASSERT(parseFlags("ce-a", "XABCDE", &result, 0, true));
    TS_ASSERT_EQUALS(result, 42);

    // correctly parse input looking like race list if it's a valid flag list
    result = 99;
    TS_ASSERT(parseFlags("xc-a", "XABCD-", &result, 0, false));
    TS_ASSERT_EQUALS(result, 43);

    /*
       now with player set
     */
    SmallSet<int> players;

    // basic operation with negative list, set initialized blank
    TS_ASSERT(parseFlags("ace-1", "XABCDE", &result, &players, false));
    TS_ASSERT_EQUALS(result, 42);
    for (int i = 1; i <= 11; ++i)
        TS_ASSERT_EQUALS(players.contains(i), (i != 1));

    // basic operation with negative list, set initialized nonblank
    players = SmallSet<int>::fromInteger(~0U);
    TS_ASSERT(parseFlags("ace-1", "XABCDE", &result, &players, true));
    TS_ASSERT_EQUALS(result, 42);
    for (int i = 1; i <= 11; ++i)
        TS_ASSERT_EQUALS(players.contains(i), (i != 1));

    // basic operation with positive list, set initialized blank
    players = SmallSet<int>::fromInteger(0);
    TS_ASSERT(parseFlags("ace+1", "XABCDE", &result, &players, true));
    TS_ASSERT_EQUALS(result, 42);
    for (int i = 1; i <= 11; ++i)
        TS_ASSERT_EQUALS(players.contains(i), (i == 1));

    // basic operation with positive list, set initialized nonblank
    players = SmallSet<int>::fromInteger(~0U);
    TS_ASSERT(parseFlags("ace+1", "XABCDE", &result, &players, false));
    TS_ASSERT_EQUALS(result, 42);
    for (int i = 1; i <= 11; ++i)
        TS_ASSERT_EQUALS(players.contains(i), (i == 1));

    // player set preserved if none specified
    players = SmallSet<int>::fromInteger(23);
    TS_ASSERT(parseFlags("ace", "XABCDE", &result, &players, false));
    TS_ASSERT_EQUALS(result, 42);
    TS_ASSERT_EQUALS(players.toInteger(), 23U);

    // player character that is part of flag set
    players = SmallSet<int>::fromInteger(0);
    TS_ASSERT(parseFlags("ce+a", "XABCDE", &result, &players, false));
    TS_ASSERT_EQUALS(result, 40);
    for (int i = 1; i <= 11; ++i)
        TS_ASSERT_EQUALS(players.contains(i), (i == 10));

    // support upper-case
    players = SmallSet<int>::fromInteger(0);
    TS_ASSERT(parseFlags("CE+A", "XABCDE", &result, &players, false));
    TS_ASSERT_EQUALS(result, 40);
    for (int i = 1; i <= 11; ++i)
        TS_ASSERT_EQUALS(players.contains(i), (i == 10));

    // refuse invalid player lists
    TS_ASSERT(!parseFlags("ce+ae", "XABCDE", &result, &players, false));
    TS_ASSERT(!parseFlags("ce+ac", "XABCDE", &result, &players, false));
    TS_ASSERT(!parseFlags("ce+a0", "XABCDE", &result, &players, false));

    // refuse invalid player lists even if mismatch allowed
    TS_ASSERT(!parseFlags("ce+ae", "XABCDE", &result, &players, true));
    TS_ASSERT(!parseFlags("ce+ac", "XABCDE", &result, &players, true));
    TS_ASSERT(!parseFlags("ce+a0", "XABCDE", &result, &players, true));
}

void
UtilMiscTestSuite::testParseRange()
{
    int min, max;
    std::size_t pos;

    // Success cases
    static const struct {
        const char* value;
        int min, max;
    } success_cases[] = {
        // standard cases
        { "42", 42, 42 },
        { "42-", 42, 100 },
        { "23-42", 23, 42 },
        { "23-142", 23, 142 },

        // various spaces
        { "  42", 42, 42 },
        { "  42-", 42, 100 },
        { "  23-42", 23, 42 },
        { "  23-142", 23, 142 },
        { "42  ", 42, 42 },
        { "42-  ", 42, 100 },
        { "23-42  ", 23, 42 },
        { "23-142  ", 23, 142 },
        { "42  -", 42, 100 },
        { "23  -42", 23, 42 },
        { "23  -142", 23, 142 },
        { "23  -  42", 23, 42 },
        { "23  -  142", 23, 142 },
        { "42  -  ", 42, 100 },
        { "23  -42  ", 23, 42 },
        { "23  -142  ", 23, 142 },
        { "23  -  42  ", 23, 42 },
        { "23  -  142  ", 23, 142 },
        { "1--2", 1, -2 }
    };

    for (unsigned i = 0; i < sizeof(success_cases)/sizeof(success_cases[0]); ++i) {
        min = 0, max = 100;
        TSM_ASSERT(success_cases[i].value, parseRange(success_cases[i].value, min, max, pos));
        TSM_ASSERT_EQUALS(success_cases[i].value, success_cases[i].min, min);
        TSM_ASSERT_EQUALS(success_cases[i].value, success_cases[i].max, max);
    }

    // Failure cases
    static const struct {
        const char* value;
        std::string::size_type pos;
    } failure_cases[] = {
        // standard failures
        { "", 0 },
        { "x", 0 },
        { "-", 0 },
        { "-2", 0 },
        { "   x", 3 },
        { "   -x", 4 },
        // { "   -2x", 5 },
        
        // standard cases
        { "42x", 2 },
        { "42-x", 3 },
        { "23-42x", 5 },

        // various spaces
        { "  42x", 4 },
        { "  42-x", 5 },
        { "  23-42x", 7 },
        { "42  x", 4 },
        { "42-  x", 5 },
        { "23-42  x", 7 },
        { "42  -x", 5 },
        { "23  -42x", 7 },
        { "23  -  42x", 9 },
        { "42  -  x", 7 },
        { "23  -42  x", 9 },
        { "23  -  42  x", 11 },
    };
    
    for (unsigned i = 0; i < sizeof(failure_cases)/sizeof(failure_cases[0]); ++i) {
        min = 0, max = 100;
        TSM_ASSERT(failure_cases[i].value, !parseRange(failure_cases[i].value, min, max, pos));
        TSM_ASSERT_EQUALS(failure_cases[i].value, failure_cases[i].pos, pos);
    }
}

void
UtilMiscTestSuite::testNumToRoman()
{
    TS_ASSERT_EQUALS(numToRoman(1), "I");
    TS_ASSERT_EQUALS(numToRoman(2), "II");
    TS_ASSERT_EQUALS(numToRoman(3), "III");
    TS_ASSERT_EQUALS(numToRoman(4), "IV");
    TS_ASSERT_EQUALS(numToRoman(5), "V");
    TS_ASSERT_EQUALS(numToRoman(6), "VI");
    TS_ASSERT_EQUALS(numToRoman(7), "VII");
    TS_ASSERT_EQUALS(numToRoman(8), "VIII");
    TS_ASSERT_EQUALS(numToRoman(9), "IX");
    TS_ASSERT_EQUALS(numToRoman(10), "X");

    TS_ASSERT_EQUALS(numToRoman(11), "XI");
    TS_ASSERT_EQUALS(numToRoman(12), "XII");
    TS_ASSERT_EQUALS(numToRoman(13), "XIII");
    TS_ASSERT_EQUALS(numToRoman(14), "XIV");
    TS_ASSERT_EQUALS(numToRoman(15), "XV");
    TS_ASSERT_EQUALS(numToRoman(16), "XVI");
    TS_ASSERT_EQUALS(numToRoman(17), "XVII");
    TS_ASSERT_EQUALS(numToRoman(18), "XVIII");
    TS_ASSERT_EQUALS(numToRoman(19), "XIX");
    TS_ASSERT_EQUALS(numToRoman(20), "XX");

    TS_ASSERT_EQUALS(numToRoman(39), "XXXIX");
    TS_ASSERT_EQUALS(numToRoman(40), "XL");
    TS_ASSERT_EQUALS(numToRoman(41), "XLI");

    TS_ASSERT_EQUALS(numToRoman(88), "LXXXVIII");
    TS_ASSERT_EQUALS(numToRoman(89), "LXXXIX");
    TS_ASSERT_EQUALS(numToRoman(90), "XC");
    TS_ASSERT_EQUALS(numToRoman(91), "XCI");
    TS_ASSERT_EQUALS(numToRoman(92), "XCII");

    TS_ASSERT_EQUALS(numToRoman(99), "XCIX");
    TS_ASSERT_EQUALS(numToRoman(100), "C");

    TS_ASSERT_EQUALS(numToRoman(1000), "M");

    TS_ASSERT_EQUALS(numToRoman(2000), "MM");
}

namespace {
    struct Counter {
        int value;
        static int count;
        Counter(const Counter&);
        Counter& operator=(const Counter&);
     public:
        Counter(int value)
            : value(value)
            { ++count; }
        ~Counter()
            { --count; }
        int getValue() const
            { return value; }
        static int getCount()
            { return count; }
    };

    int Counter::count;
}

void
UtilMiscTestSuite::testSafeAppendClear()
{
    // safeAppend(), safeClear(). Since we cannot easily trigger exceptions, just basic tests here.

    // Make a vector:
    Counter* nil = 0;
    std::vector<Counter*> vec;
    TS_ASSERT_EQUALS(Counter::getCount(), 0);
    safeAppend(vec, new Counter(47));
    safeAppend(vec, new Counter(11));
    TS_ASSERT_EQUALS(Counter::getCount(), 2);
    TS_ASSERT_EQUALS(vec.size(), 2U);

    TS_ASSERT_DIFFERS(vec[0], nil); /* nil needed for type-correctness in TS_ASSERT_DIFFERS here */
    TS_ASSERT_DIFFERS(vec[1], nil);
    TS_ASSERT_DIFFERS(vec[0], vec[1]);
    TS_ASSERT_EQUALS(vec[0]->getValue(), 47);
    TS_ASSERT_EQUALS(vec[1]->getValue(), 11);

    // Destroy it:
    safeClear(vec);
    TS_ASSERT_EQUALS(Counter::getCount(), 0);
    TS_ASSERT_EQUALS(vec.size(), 2U);
    TS_ASSERT_EQUALS(vec[0], nil);
    TS_ASSERT_EQUALS(vec[1], nil);
}
